#include "adu.h"
#include <dirent.h> /* readdir() */

#include "gcc-compat.h"
#include "osl.h"
#include "fd.h"
#include "hash.h"
#include "string.h"
#include "error.h"

DEFINE_ERRLIST;

/** evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))


/**
 * The log function.
 *
 * \param ll Loglevel.
 * \param fml Usual format string.
 *
 * All XXX_LOG() macros use this function.
 */
__printf_2_3 void __log(int ll, const char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[255] = "";

	if (ll < 4)
		return;
	outfd = stderr;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, sizeof(str), "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

/**
 * Compare the size of two directories
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * This function first compares the size values as usual integers. If they compare as
 * equal, the address of \a obj1 and \a obj2 are compared. So this compare function
 * returns zero if and only if \a obj1 and \a obj2 point to the same memory area.
 */
static int size_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	uint64_t d1 = *(uint64_t *)obj1->data;
	uint64_t d2 = *(uint64_t *)obj2->data;
	int ret = NUM_COMPARE(d2, d1);

	if (ret)
		return ret;
	//INFO_LOG("addresses: %p, %p\n", obj1->data, obj2->data);
	return NUM_COMPARE(obj2->data, obj1->data);
}

/**
 * Compare two osl objects of string type.
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * In any case, only \p MIN(obj1->size, obj2->size) characters of each string
 * are taken into account.
 *
 * \return It returns an integer less than, equal to, or greater than zero if
 * \a obj1 is found, respectively, to be less than, to match, or be greater than
 * obj2.
 *
 * \sa strcmp(3), strncmp(3), osl_compare_func.
 */
int string_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	const char *str1 = (const char *)obj1->data;
	const char *str2 = (const char *)obj2->data;
	return strncmp(str1, str2, MIN(obj1->size, obj2->size));
}

/** The columns of the directory table. */
enum dir_table_columns {
	/** The name of the directory. */
	DT_NAME,
	/** The hash value of the name. */
	DT_HASH,
	/** The number of bytes of all regular files. */
	DT_BYTES,
	/** The number of all regular files. */
	DT_FILES,
	/** Number of columns in this table. */
	NUM_DT_COLUMNS
};

static struct osl_column_description dir_table_cols[] = {
	[DT_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE,
		.name = "dir",
		.compare_function = string_compare,
	},
	[DT_HASH] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "hash",
		.compare_function = osl_hash_compare,
		.data_size = HASH_SIZE
	},
	[DT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[DT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags =  OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	}
};

static struct osl_table_description dir_table_desc = {
	.name = "dir_table",
	.num_columns = NUM_DT_COLUMNS,
	.flags = 0,
	.column_descriptions = dir_table_cols,
	.dir = "/tmp/adu"
};

/** The columns of the id table. */
enum id_table_columns {
	/** The user id. */
	IDT_UID,
	/** The number of bytes of all regular files owned by this id. */
	IDT_BYTES,
	/** The number of regular files owned by this id. */
	IDT_FILES,
	/** The user table for this uid. */
	IDT_TABLE,
	/** Number of columns in this table. */
	NUM_IDT_COLUMNS
};

static struct osl_column_description id_table_cols[] = {
	[IDT_UID] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "uid",
		.compare_function = uint32_compare,
		.data_size = sizeof(uint32_t)
	},
	[IDT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[IDT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_filess",
		.data_size = sizeof(uint64_t)
	},
	[IDT_TABLE] = {
		.storage_type = OSL_NO_STORAGE,
		.storage_flags = OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "user_table",
		.data_size = sizeof(void *)
	}
};

static struct osl_table_description id_table_desc = {
	.name = "id_table",
	.num_columns = NUM_IDT_COLUMNS,
	.flags = 0,
	.column_descriptions = id_table_cols,
	.dir = "/tmp/adu"
};

/** The columns of the id table. */
enum user_table_columns {
	/** The hash of the directory. */
	UT_DIR_HASH,
	/** The number of bytes of all regular files in this dir owned by this id. */
	UT_BYTES,
	/** The number of files in this dir owned by this id. */
	UT_FILES,
	/** Number of columns in this table. */
	NUM_UT_COLUMNS
};

static struct osl_column_description user_table_cols[] = {
	[UT_DIR_HASH] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "dir_hash",
		.compare_function = osl_hash_compare,
		.data_size = HASH_SIZE
	},
	[IDT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[IDT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	},
};

static struct osl_table_description user_table_desc = {
	.num_columns = NUM_UT_COLUMNS,
	.flags = 0,
	.column_descriptions = user_table_cols,
	.dir = "/tmp/adu"
};
static struct osl_table *dir_table;
static struct osl_table *id_table;

static int create_tables(void)
{
	int ret = osl_create_table(&dir_table_desc);
	if (ret < 0)
		return ret;
	ret = osl_create_table(&id_table_desc);
	if (ret < 0)
		return ret;
	return 1;
}

int add_directory(char *dirname, uint64_t *dir_size, uint64_t *dir_files)
{
	struct osl_object dir_objects[NUM_DT_COLUMNS];
	HASH_TYPE hash[HASH_SIZE];

	INFO_LOG("scanning %s\n", dirname);
	dir_objects[DT_NAME].data = dirname;
	dir_objects[DT_NAME].size = strlen(dirname) + 1;
	hash_function(dirname, strlen(dirname), hash);
	dir_objects[DT_HASH].data = hash;
	dir_objects[DT_HASH].size = HASH_SIZE;
	dir_objects[DT_BYTES].data = dir_size;
	dir_objects[DT_BYTES].size = sizeof(*dir_size);
	dir_objects[DT_FILES].data = dir_files;
	dir_objects[DT_FILES].size = sizeof(*dir_files);

	return osl_add_row(dir_table, dir_objects);
}

int create_and_open_user_table(uint32_t uid, struct osl_table **t)
{
	int ret;
	struct osl_table_description *desc = para_malloc(sizeof(*desc));

	desc->num_columns = NUM_UT_COLUMNS;
	desc->flags = 0;
	desc->column_descriptions = user_table_cols;
	desc->dir = para_strdup("/tmp/adu");
	desc->name = make_message("%u", uid);
	INFO_LOG("................................. %u\n", uid);
//	user_table_desc.name = make_message("%u", uid);
	ret = osl_create_table(desc);
	if (ret < 0)
		return ret;
	return osl_open_table(desc, t);
}

static int insert_id_row(uint32_t uid, struct osl_table *t, struct osl_row **row)
{
	struct osl_object objects[NUM_IDT_COLUMNS];
	uint64_t num = 0;

	struct osl_table **table_ptr = para_malloc(sizeof(*table_ptr));
	*table_ptr = t;

	INFO_LOG(" uid: %d, t: %p\n", uid, t);
	objects[IDT_UID].data = &uid;
	objects[IDT_UID].size = sizeof(uid);
	objects[IDT_BYTES].data = &num;
	objects[IDT_BYTES].size = sizeof(num);
	objects[IDT_FILES].data = &num;
	objects[IDT_FILES].size = sizeof(num);
	objects[IDT_TABLE].data = table_ptr;
	objects[IDT_TABLE].size = sizeof(*table_ptr);
	return osl_add_and_get_row(id_table, objects, row);
}

static int get_user_table(struct osl_row *row, struct osl_table **t)
{
	struct osl_object obj;

	int ret = osl_get_object(id_table, row, IDT_TABLE, &obj);
	if (ret < 0)
		return ret;
	*t = *(struct osl_table **)obj.data;
	INFO_LOG("^^^^^^^^^^^^^^^^^^ t: %p\n", *t);
	return 1;
}

static int add_id_bytes(struct osl_row *row, uint64_t *add)
{
	uint64_t num;
	struct osl_object obj1, obj2 = {.data = &num, .size = sizeof(num)};

	/* update number of bytes */
	int ret = osl_get_object(id_table, row, IDT_BYTES, &obj1);
	if (ret < 0)
		return ret;
	num = *(uint64_t *)obj1.data + *add;
	ret = osl_update_object(id_table, row, IDT_BYTES, &obj2);
	if (ret < 0)
		return ret;
	/* increment number of files */
	ret = osl_get_object(id_table, row, IDT_FILES, &obj1);
	if (ret < 0)
		return ret;
	num = *(uint64_t *)obj1.data + 1;
	return osl_update_object(id_table, row, IDT_FILES, &obj2);
}

static int update_user_row(struct osl_table *t, HASH_TYPE *hash,
		uint64_t *add)
{
	struct osl_row *row;
	struct osl_object obj = {.data = hash, .size = HASH_SIZE};

	int ret = osl_get_row(t, UT_DIR_HASH, &obj, &row);

	if (ret < 0 && ret != -E_RB_KEY_NOT_FOUND)
		return ret;
	if (ret < 0) { /* this is the first file we add */
		struct osl_object objects[NUM_UT_COLUMNS];
		uint64_t num_files = 1;

		objects[UT_DIR_HASH].data = hash;
		objects[UT_DIR_HASH].size = HASH_SIZE;
		objects[UT_BYTES].data = add;
		objects[UT_BYTES].size = sizeof(*add);
		objects[UT_FILES].data = &num_files;
		objects[UT_FILES].size = sizeof(num_files);
		INFO_LOG("######################### ret: %d\n", ret);
		ret = osl_add_row(t, objects);
		INFO_LOG("######################### ret: %d\n", ret);
		return ret;
	} else { /* add size and increment file count */
		uint64_t num;
		struct osl_object obj1, obj2 = {.data = &num, .size = sizeof(num)};

		ret = osl_get_object(t, row, UT_BYTES, &obj1);
		if (ret < 0)
			return ret;
		num = *(uint64_t *)obj1.data + *add;
		ret = osl_update_object(t, row, UT_BYTES, &obj2);
		if (ret < 0)
			return ret;
		ret = osl_get_object(t, row, UT_FILES, &obj1);
		if (ret < 0)
			return ret;
		num = *(uint64_t *)obj1.data + 1;
		return osl_update_object(t, row, UT_FILES, &obj2);
	}
}

int scan_dir(char *dirname)
{
	DIR *dir;
	struct dirent *entry;
	int ret, cwd_fd, ret2;
	uint64_t dir_size = 0, dir_files = 0;
	struct osl_object obj;
	HASH_TYPE hash[HASH_SIZE];

	INFO_LOG("----------------- %s\n", dirname);
	ret = para_opendir(dirname, &dir, &cwd_fd);
	if (ret < 0) {
		if (ret != -ERRNO_TO_ERROR(EACCES))
			return ret;
		WARNING_LOG("permission denied for %s\n", dirname);
		return 1;
	}
	hash_function(dirname, strlen(dirname), hash);
	while ((entry = readdir(dir))) {
		mode_t m;
		char *tmp;
		struct stat s;
		uint32_t uid;
		uint64_t size;
		struct osl_row *id_row;
		struct osl_table *user_table;

		if (!strcmp(entry->d_name, "."))
			continue;
		if (!strcmp(entry->d_name, ".."))
			continue;
		if (lstat(entry->d_name, &s) == -1)
			continue;
		m = s.st_mode;
		if (!S_ISREG(m) && !S_ISDIR(m))
			continue;
		tmp = make_message("%s/%s", dirname, entry->d_name);
		if (S_ISDIR(m)) {
			ret = scan_dir(tmp);
			free(tmp);
			if (ret < 0)
				goto out;
			continue;
		}
		/* regular file */
		size = s.st_size;
		dir_size += size;
		dir_files++;
		uid = s.st_uid;
		INFO_LOG("++++++++++++++++++++++++++ %s, uid: %u\n", entry->d_name, uid);
		obj.data = &uid;
		obj.size = sizeof(uid);
		ret = osl_get_row(id_table, IDT_UID, &obj, &id_row);
		if (ret < 0 && ret != -E_RB_KEY_NOT_FOUND)
			goto out;
		if (ret < 0) {
			ret = create_and_open_user_table(uid, &user_table);
			if (ret < 0)
				goto out;
			ret = insert_id_row(uid, user_table, &id_row);
			if (ret < 0)
				goto out;
		} else {
			ret = get_user_table(id_row, &user_table);
			if (ret < 0)
				goto out;
		}
		ret = add_id_bytes(id_row, &size);
		if (ret < 0)
			goto out;
		INFO_LOG("user_table: %p\n", user_table);
		ret = update_user_row(user_table, hash, &size);
		INFO_LOG("update_user  ret: %d\n", ret);
		if (ret < 0)
			goto out;
	}
	ret = add_directory(dirname, &dir_size, &dir_files);
out:
	closedir(dir);
	ret2 = para_fchdir(cwd_fd);
	if (ret2 < 0 && ret >= 0)
		ret = ret2;
	close(cwd_fd);
	return ret;
}

static int get_dir_name(struct osl_row *row, char **name)
{
	struct osl_object obj;
	int ret = osl_get_object(dir_table, row, DT_NAME, &obj);

	if (ret < 0)
		return ret;
	*name = obj.data;
	return 1;
}

static int print_dirname_and_size(struct osl_row *row, void *data)
{
	unsigned *count = data;
	struct osl_object obj;
	char *name;
	int ret;

	if ((*count)++ > 100)
		return -E_LOOP_COMPLETE;
	ret = get_dir_name(row, &name);
	if (ret < 0)
		return ret;
	ret = osl_get_object(dir_table, row, DT_BYTES, &obj);
	if (ret < 0)
		return ret;
	printf("%s\t%llu\n", name, *(long long unsigned *)obj.data);
	return 1;
}

static int print_dirname_and_file_count(struct osl_row *row, void *data)
{
	unsigned *count = data;
	struct osl_object obj;
	char *name;
	int ret;

	if ((*count)++ > 100)
		return -E_LOOP_COMPLETE;
	ret = get_dir_name(row, &name);
	if (ret < 0)
		return ret;
	ret = osl_get_object(dir_table, row, DT_FILES, &obj);
	if (ret < 0)
		return ret;
	printf("%s\t%llu\n", name, *(long long unsigned *)obj.data);
	return 1;
}

static int print_id_stats(struct osl_row *row, __a_unused void *data)
{
	struct osl_object obj;
	uint32_t uid;
	uint64_t bytes, files;
	int ret = osl_get_object(id_table, row, IDT_UID, &obj);

	if (ret < 0)
		return ret;
	uid = *(uint32_t *)obj.data;
	ret = osl_get_object(id_table, row, IDT_BYTES, &obj);
	if (ret < 0)
		return ret;
	bytes = *(uint64_t *)obj.data;
	ret = osl_get_object(id_table, row, IDT_FILES, &obj);
	if (ret < 0)
		return ret;
	files = *(uint64_t *)obj.data;

	printf("%u\t%llu%llu\n", (unsigned)uid, (long long unsigned)files,
		(long long unsigned)bytes);
	return 1;
}

struct id_dir_stat_info {
	unsigned count;
	struct osl_table *user_table;
};

static int print_big_dir(struct osl_row *row, void *data)
{
	struct id_dir_stat_info *info = data;
	info->count++;
	int ret;
	struct osl_row *dir_row;
	char *dirname;
	uint64_t bytes;
	struct osl_object obj;

	if (info->count > 10)
		return -E_LOOP_COMPLETE;
	ret = osl_get_object(info->user_table, row, UT_BYTES, &obj);
	if (ret < 0)
		return ret;
	bytes = *(uint64_t *)obj.data;
	ret = osl_get_object(info->user_table, row, UT_DIR_HASH, &obj);
	if (ret < 0)
		return ret;
	ret = osl_get_row(dir_table, DT_HASH, &obj, &dir_row);
	if (ret < 0)
		return ret;
	ret = osl_get_object(dir_table, dir_row, DT_NAME, &obj);
	if (ret < 0)
		return ret;
	dirname = obj.data;
	printf("%s: %llu\n", dirname, (long long unsigned)bytes);
	return 1;
}

static int print_id_dir_stats(struct osl_row *row, __a_unused void *data)
{
	struct osl_object obj;
	uint32_t uid;
	int ret = osl_get_object(id_table, row, IDT_UID, &obj);
	struct id_dir_stat_info info = {.count = 0};

	if (ret < 0)
		return ret;
	uid = *(uint32_t *)obj.data;

	ret = osl_get_object(id_table, row, IDT_TABLE, &obj);
	if (ret < 0)
		return ret;
	info.user_table = *(struct osl_table **)obj.data;

	printf("************************* Big dirs owned by uid %u\n", (unsigned) uid);
	osl_rbtree_loop_reverse(info.user_table, IDT_BYTES, &info, print_big_dir);
	return 1;
}

static int print_statistics(void)
{
	unsigned count = 0;
	int ret;

	printf("************************* Biggest dirs\n");
	ret = osl_rbtree_loop_reverse(dir_table, DT_BYTES, &count, print_dirname_and_size);
	if (ret < 0 && ret != -E_LOOP_COMPLETE)
		return ret;
	count = 0;
	printf("************************* dirs containing many files\n");
	ret = osl_rbtree_loop_reverse(dir_table, DT_FILES, &count, print_dirname_and_file_count);
	if (ret < 0 && ret != -E_LOOP_COMPLETE)
		return ret;

	printf("************************* dirs stats by owner\n");
	ret = osl_rbtree_loop(id_table, IDT_BYTES, NULL, print_id_stats);
	if (ret < 0)
		return ret;

	return osl_rbtree_loop(id_table, IDT_BYTES, NULL, print_id_dir_stats);
}


int main(int argc, char **argv)
{
	int ret = create_tables();
	if (ret < 0)
		goto out;
	ret = osl_open_table(&dir_table_desc, &dir_table);
	if (ret < 0)
		goto out;
	ret = osl_open_table(&id_table_desc, &id_table);
	if (ret < 0)
		goto out;
	ret = -E_SYNTAX;
	if (argc != 2)
		goto out;
	ret = scan_dir(argv[1]);
	if (ret < 0)
		goto out;
	print_statistics();
out:
	if (ret < 0) {
		ERROR_LOG("%s\n", error_txt(-ret));
		return -EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

